from mite_extras.cli.cli_manager import CliManager
from mite_extras.processing.file_manager import FileManager
from mite_extras.processing.mite_parser import MiteParser

__all__ = [
    "CliManager",
    "FileManager",
    "MiteParser",
]



"""Validation and internal representation of MITE schema.

Copyright (c) 2024 to present Mitja Maximilian Zdouc, PhD and individual contributors.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import base64
import logging
import re
from typing import Any, Self

from pydantic import BaseModel, ValidationError, model_validator
from rdkit.Chem import MolFromSmiles
from rdkit.Chem.Draw import rdMolDraw2D
from rdkit.Chem.rdChemReactions import ReactionFromSmarts

from mite_extras.processing.validation_manager import (
    IdValidator,
    MoleculeValidator,
    ReactionValidator,
)

logger = logging.getLogger("mite_extras")


class Entry(BaseModel):
    """Pydantic-based class to represent a MITE entry

    Attributes:
        accession: MITE accession number/identifier.
        status: status of entry
        retirementReasons: a list of retirement reasons
        changelog:  a list of Changelog objects
        enzyme: an Enzyme object
        reactions: a list of Reaction objects
        comment: additional information
    """

    accession: str | None = None
    status: str | None = None
    retirementReasons: list[str] | None = None
    changelog: list[Any] | None = None
    enzyme: Any | None = None
    reactions: list[Any] | None = None
    comment: str | None = None

    def to_json(self: Self) -> dict:
        json_dict = {}
        for attr in ["accession", "status", "retirementReasons", "comment"]:
            val = getattr(self, attr)
            if val not in (None, ""):
                json_dict[attr] = val

        if self.changelog:
            json_dict["changelog"] = [
                changelog.to_json() for changelog in self.changelog
            ]

        if self.enzyme:
            json_dict["enzyme"] = self.enzyme.to_json()

        if self.reactions:
            json_dict["reactions"] = [reaction.to_json() for reaction in self.reactions]

        return json_dict

    def to_html(self: Self) -> dict:
        html_dict = {}
        for attr in ["accession", "status", "retirementReasons", "comment"]:
            val = getattr(self, attr)
            if val not in (None, ""):
                html_dict[attr] = val

        if self.changelog:
            html_dict["changelog"] = [
                changelog.to_html() for changelog in self.changelog
            ]

        if self.enzyme:
            html_dict["enzyme"] = self.enzyme.to_html()

        if self.reactions:
            html_dict["reactions"] = [reaction.to_html() for reaction in self.reactions]

        return html_dict


class Changelog(BaseModel):
    """Pydantic-based class to represent changelog

    Attributes:
        version: the MITE version
        contributors: a list of contributors
        reviewers: a list of reviewers
        date: the release date in YYYY-MM-DD
        comment: comment indicating changes
    """

    version: str
    contributors: list
    reviewers: list
    date: str
    comment: str

    @model_validator(mode="after")
    def check_empty(self):
        if self.comment == "":
            raise ValidationError("Changelog: comment must not be an empty string.")
        return self

    def to_json(self: Self) -> dict:
        return {
            attr: getattr(self, attr)
            for attr in ["version", "date", "contributors", "reviewers", "comment"]
        }

    def to_html(self: Self) -> dict:
        return self.to_json()


class Enzyme(BaseModel):
    """Pydantic-based class to represent enzyme information

    Attributes:
        name: the protein name
        description: a brief description of the enzyme
        databaseIds: an EnzymeDatabaseIds object
        auxiliaryEnzymes: a list of EnzymeAux objects
        references: a list of references
        cofactors: optional object containing cofactor lists
    """

    name: str
    description: str | None = None
    databaseIds: Any
    auxiliaryEnzymes: list | None = None
    references: list
    cofactors: Any | None = None

    def to_json(self: Self) -> dict:
        json_dict = {}
        for attr in ["name", "description", "references"]:
            val = getattr(self, attr)
            if val not in (None, ""):
                json_dict[attr] = val

        if self.auxiliaryEnzymes:
            json_dict["auxiliaryEnzymes"] = [
                entry.to_json() for entry in self.auxiliaryEnzymes
            ]

        if self.databaseIds.to_json() == {}:
            raise RuntimeError(
                "At least one of 'uniprot' and 'genpept' IDs must be provided."
            )
        else:
            json_dict["databaseIds"] = self.databaseIds.to_json()

        if self.cofactors and self.cofactors.to_json():
            json_dict["cofactors"] = self.cofactors.to_json()

        return json_dict

    def to_html(self: Self) -> dict:
        html_dict = {}
        for attr in ["name", "description", "references"]:
            val = getattr(self, attr)
            if val not in (None, ""):
                html_dict[attr] = val

        if self.auxiliaryEnzymes:
            html_dict["auxiliaryEnzymes"] = [
                entry.to_html() for entry in self.auxiliaryEnzymes
            ]

        html_dict["databaseIds"] = self.databaseIds.to_html()

        if self.cofactors and self.cofactors.to_html():
            html_dict["cofactors"] = self.cofactors.to_html()

        return html_dict


class EnzymeAux(BaseModel):
    """Pydantic-based class to represent auxiliary enzyme information

    Attributes:
        name: name of the auxiliary enzyme
        description: an optional description
        databaseIds: an EnyzmeDatabaseIds instance
    """

    name: str
    description: str | None = None
    databaseIds: Any

    def to_json(self: Self) -> dict:
        json_dict = {}
        for attr in ["name", "description"]:
            val = getattr(self, attr)
            if val not in (None, ""):
                json_dict[attr] = val

        if self.databaseIds.to_json() != {}:
            json_dict["databaseIds"] = self.databaseIds.to_json()

        return json_dict

    def to_html(self: Self) -> dict:
        html_dict = {}
        for attr in ["name", "description"]:
            val = getattr(self, attr)
            if val not in (None, ""):
                html_dict[attr] = val

        html_dict["databaseIds"] = self.databaseIds.to_html()

        return html_dict


class EnyzmeDatabaseIds(BaseModel):
    """Pydantic-based class to represent enzyme-related database ids

    Class-level instance of IdValidator avoids repeated instantiation

    Attributes:
        uniprot: a UniProt ID
        genpept: an NCBI GenPept ID
        mibig: a MIBiG ID
        wikidata: a wikidata qid reference
    """

    uniprot: str | None = None
    genpept: str | None = None
    mibig: str | None = None
    wikidata: str | None = None

    _id_validator = IdValidator()

    @model_validator(mode="after")
    def validate_ids(self):
        """Cross-reference and populate database IDs using IdValidator.

        Checks if genpept and uniprot IDs correspond
        If only one ID is provided, fetches the other
        Checks if wikidata QID exists

        Raises:
            ValueError: did not provide one of uniprot or genpept IDs
        """
        if not self.genpept and not self.uniprot:
            raise ValueError("Neither 'UniProt' not 'GenPept' IDs provided.")

        try:
            if self.uniprot and self.genpept:
                self._id_validator.cleanup_ids(
                    genpept=self.genpept, uniprot=self.uniprot
                )
            elif self.uniprot:
                data = self._id_validator.cleanup_ids(uniprot=self.uniprot)
                self.genpept = data["genpept"]
            elif self.genpept:
                data = self._id_validator.cleanup_ids(genpept=self.genpept)
                self.uniprot = data["uniprot"]
        except Exception as e:
            logger.warning(f"EnyzmeDatabaseIds: error during ID validation: {e!s}")

        try:
            if self.wikidata:
                self._id_validator.validate_wikidata_qid(self.wikidata)
        except Exception as e:
            logger.error(f"EnyzmeDatabaseIds: error during ID validation: {e!s}")
            raise e from Exception

        return self

    def to_json(self: Self) -> dict:
        json_dict = {}
        for attr in ["uniprot", "genpept", "mibig", "wikidata"]:
            val = getattr(self, attr)
            if val not in (None, ""):
                json_dict[attr] = val
        return json_dict

    def to_html(self: Self) -> dict:
        return self.to_json()


class Reaction(BaseModel):
    """Pydantic-based class to represent reactions

    Class-level instance of ReactionValidator avoids repeated instantiation

    Attributes:
        tailoring: a list of tailoring reaction terms
        description: an optional human-readable description
        reactionSMARTS: a ReactionSmarts str
        reactions: a list of ReactionEx objects
        evidence: an evidence object
        databaseIds: a ReactionDatabaseIds object
    """

    tailoring: list
    description: str | None = None
    reactionSMARTS: str
    reactions: list
    evidence: Any
    databaseIds: Any | None = None

    _reaction_validator = ReactionValidator()

    @model_validator(mode="after")
    def cleanup_smarts(self):
        """Clean up reaction SMARTS using the new ReactionCleaner

        Raises:
            ValueError: reaction SMARTS could not be read by RDKit -> invalid
        """
        try:
            cleaned_smarts = (
                self._reaction_validator.reaction_cleaner.clean_ketcher_format(
                    self._reaction_validator.molecule_validator._clean_string(
                        self.reactionSMARTS
                    )
                )
            )

            if ReactionFromSmarts(cleaned_smarts) is None:
                raise ValueError(f"Invalid reaction SMARTS: {self.reactionSMARTS}")

            self.reactionSMARTS = cleaned_smarts
            return self

        except Exception as e:
            raise ValueError(f"Error cleaning reaction SMARTS: {e!s}") from e

    @model_validator(mode="after")
    def validate_reactions(self):
        """Validate all reactions using the new ReactionValidator

        Raises:
            ValueError: Reaction validation failed
        """
        for reaction_id, reaction in enumerate(self.reactions, start=1):
            intramolecular = False
            if re.match(r"^\(.+\)>>|>>\(.+\)$", self.reactionSMARTS):
                intramolecular = True

            try:
                self._reaction_validator.validate_reaction(
                    reaction_smarts=self.reactionSMARTS,
                    substrate_smiles=reaction.substrate,
                    expected_products=reaction.products,
                    forbidden_products=reaction.forbidden_products,
                    intramolecular=intramolecular,
                )
            except Exception as e:
                raise ValueError(
                    f"Reaction '{reaction_id}': Validation failed for substrate {reaction.substrate}: {e!s}"
                ) from e
        return self

    def to_json(self: Self) -> dict:
        json_dict = {}

        for attr in ["tailoring", "description"]:
            val = getattr(self, attr)
            if val not in (None, ""):
                json_dict[attr] = val

        json_dict["reactionSMARTS"] = self.reactionSMARTS
        json_dict["reactions"] = [entry.to_json() for entry in self.reactions]
        json_dict["evidence"] = self.evidence.to_json()

        if self.databaseIds and self.databaseIds.to_json() != {}:
            json_dict["databaseIds"] = self.databaseIds.to_json()

        return json_dict

    def to_html(self: Self) -> dict:
        def _smarts_to_svg(smarts: str) -> str:
            """Generates a base64 encoded SVG string of the reaction SMARTS"""
            rxn = ReactionFromSmarts(smarts)
            drawer = rdMolDraw2D.MolDraw2DSVG(-1, -1)
            dopts = drawer.drawOptions()
            dopts.padding = 1e-5
            dopts.clearBackground = False
            drawer.DrawReaction(
                rxn,
                highlightByReactant=True,
                highlightColorsReactants=[(0.69, 0.863, 0.949)],  # RGB blue
            )
            drawer.FinishDrawing()

            svg = drawer.GetDrawingText()
            return base64.b64encode(svg.encode("utf-8")).decode("utf-8")

        html_dict = {}

        for attr in ["tailoring", "description"]:
            val = getattr(self, attr)
            if val not in (None, ""):
                html_dict[attr] = val

        html_dict["reactionSMARTS"] = (
            self.reactionSMARTS,
            _smarts_to_svg(self.reactionSMARTS),
        )
        html_dict["reactions"] = [entry.to_html() for entry in self.reactions]

        evidences = self.evidence.to_html()
        html_dict["evidenceCode"] = evidences["evidenceCode"]
        html_dict["references"] = evidences["references"]

        if self.databaseIds:
            html_dict["databaseIds"] = self.databaseIds.to_html()

        return html_dict


class ReactionEx(BaseModel):
    """Pydantic-based class to represent data of experimentally verified reaction

    Attributes:
        substrate: SMILES string
        products: list of SMILES strings (result from substrate if reactionSMARTS appl)
        forbidden_products: list of SMILES string of forbidden products (must not result)
        isIntermediate: is the reaction an intermediate or a stable product
        description: an optional string
    """

    substrate: str
    products: list
    forbidden_products: list | None = None
    isIntermediate: bool
    description: str | None = None

    # Create validator instance at class level
    _molecule_validator = MoleculeValidator()

    @model_validator(mode="after")
    def validate_smiles(self):
        """Validate and clean SMILES using the new MoleculeValidator"""
        try:
            # Clean substrate SMILES
            self.substrate = self._molecule_validator.canonicalize_smiles(
                self._molecule_validator._clean_string(self.substrate)
            )

            # Clean product SMILES
            self.products = [
                self._molecule_validator.canonicalize_smiles(
                    self._molecule_validator._clean_string(prod)
                )
                for prod in self.products
            ]

            # Clean forbidden product SMILES if present
            if self.forbidden_products:
                cleaned_forbidden = []
                for prod in self.forbidden_products:
                    cleaned = self._molecule_validator.canonicalize_smiles(
                        self._molecule_validator._clean_string(prod)
                    )
                    # Split composite SMILES into individual molecules
                    cleaned_forbidden.extend(cleaned.split("."))
                self.forbidden_products = cleaned_forbidden

            return self

        except Exception as e:
            raise ValueError(f"Error during SMILES cleanup: {e!s}") from e

    def to_json(self: Self) -> dict:
        json_dict = {}

        for attr in [
            "substrate",
            "products",
            "forbidden_products",
            "isIntermediate",
            "description",
        ]:
            val = getattr(self, attr)
            if val not in (None, ""):
                json_dict[attr] = val

        return json_dict

    def to_html(self: Self) -> dict:
        def _smiles_to_svg(smiles: str) -> str:
            """Generates a base64 encoded SVG strin"""
            m = MolFromSmiles(smiles)

            for atom in m.GetAtoms():
                atom.SetAtomMapNum(0)

            m = rdMolDraw2D.PrepareMolForDrawing(m)

            drawer = rdMolDraw2D.MolDraw2DSVG(-1, -1)
            dopts = drawer.drawOptions()
            dopts.clearBackground = False

            drawer.DrawMolecule(m)
            drawer.FinishDrawing()
            svg = drawer.GetDrawingText()
            return base64.b64encode(svg.encode("utf-8")).decode("utf-8")

        html_dict = {}

        for attr in [
            "isIntermediate",
            "description",
        ]:
            val = getattr(self, attr)
            if val not in (None, ""):
                html_dict[attr] = val

        html_dict["substrate"] = (self.substrate, _smiles_to_svg(self.substrate))

        html_dict["products"] = [(mol, _smiles_to_svg(mol)) for mol in self.products]

        if self.forbidden_products:
            html_dict["forbidden_products"] = [
                (mol, _smiles_to_svg(mol)) for mol in self.forbidden_products
            ]

        return html_dict


class ReactionDatabaseIds(BaseModel):
    """Pydantic-based class to represent reaction-related database ids

    Attributes:
        rhea: a RHEA ID
        ec: an EC (Enzyme Commission) number
    """

    rhea: str | None = None
    ec: str | None = None

    def to_json(self: Self) -> dict:
        json_dict = {}
        for attr in ["rhea", "ec"]:
            val = getattr(self, attr)
            if val not in (None, ""):
                json_dict[attr] = val
        return json_dict

    def to_html(self: Self) -> dict:
        return self.to_json()


class Evidence(BaseModel):
    """Pydantic-based class to represent evidence information

    Attributes:
        evidenceCode: a list of evidence code strings
        references: a list of references
    """

    evidenceCode: list
    references: list

    def to_json(self: Self) -> dict:
        return {"evidenceCode": self.evidenceCode, "references": self.references}

    def to_html(self: Self) -> dict:
        return self.to_json()


class Cofactors(BaseModel):
    """Pydantic-based class representing cofactor information

    Attributes:
        organic: a list of organic cofactors
        inorganic: a list of inorganic cofactors
    """

    organic: list | None = None
    inorganic: list | None = None

    def to_json(self: Self) -> dict:
        json_dict = {}
        for attr in ["organic", "inorganic"]:
            val = getattr(self, attr)
            if val not in (None, []):
                json_dict[attr] = val
        return json_dict

    def to_html(self: Self) -> dict:
        return self.to_json()



"""Main entry point to mite_extras.

Copyright (c) 2024 to present Mitja M. Zdouc and individual contributors.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging
import sys
from importlib import metadata

import coloredlogs
import json_repair
from mite_schema import SchemaManager

from mite_extras import CliManager, FileManager, MiteParser


def config_logger(verboseness: str) -> logging.Logger:
    """Set up a named logger with nice formatting

    Args:
        verboseness: sets the logging verboseness

    Returns:
        A Logger object
    """
    logger = logging.getLogger("mite_extras")
    logger.setLevel(getattr(logging, verboseness))
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(
        coloredlogs.ColoredFormatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
    )
    logger.addHandler(console_handler)
    return logger


def main_cli() -> None:
    """Entry point for CLI"""
    args = CliManager().run(sys.argv[1:])

    logger = config_logger(args.verboseness)
    logger.debug(f"Started 'mite_extras' v{metadata.version('mite_extras')} as CLI.")

    schema_manager = SchemaManager()

    file_manager = FileManager(indir=args.input_dir, outdir=args.output_dir)
    file_manager.read_files_indir()

    for entry in file_manager.infiles:
        logger.info(f"CLI: started parsing of file '{entry.name}'.")

        with open(entry) as infile:
            input_data = json_repair.load(infile)

        try:
            parser = MiteParser()
            parser.parse_mite_json(data=input_data)

            schema_manager.validate_mite(instance=parser.to_json())

            file_manager.write_json(outfile_name=entry.stem, payload=parser.to_json())

            logger.info(f"CLI: completed parsing of file '{entry.name}'.")
        except Exception as e:
            logger.fatal(f"Could not process file '{entry.name}': {e!s}")
            continue

    logger.info("Completed 'mite_extras' as CLI.")


if __name__ == "__main__":
    main_cli()



"""Validates and stores input and output file references

Copyright (c) 2024 to present Mitja Maximilian Zdouc, PhD and individual contributors.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
import logging
from pathlib import Path
from typing import Self

from pydantic import BaseModel, model_validator

logger = logging.getLogger("mite_extras")


class FileManager(BaseModel):
    """Pydantic-based class to validate and store input/output file references

    Attributes:
        indir: input directory path
        infiles: input files paths
        outdir: output directory path
    """

    indir: Path
    infiles: list = []
    outdir: Path

    @model_validator(mode="after")
    def validate_exists(self):
        """Validate existence of input/output dirs

        Raises:
            FileNotFoundError: one or more dirs could not be found
        """
        for path in [self.indir, self.outdir]:
            if not path.exists():
                logger.fatal(f"FileManager: could not find directory '{path.name}'")
                raise FileNotFoundError
        return self

    def read_files_indir(self: Self) -> None:
        """Read files in the indir, store to self"""
        logger.debug(
            f"FileManager: started reading file in input directory '{self.indir.name}'."
        )

        for infile in self.indir.iterdir():
            if infile.suffix != ".json":
                logger.warning(
                    f"FileManager: file '{infile.name}' does not apprear to be in .json-format - SKIP."
                )
                continue
            self.infiles.append(infile)

        logger.debug(
            f"FileManager: completed reading file in input directory '{self.indir.name}'."
        )

    def write_json(self: Self, outfile_name: str, payload: dict) -> None:
        """Write dict as json file to outdir

        Args:
            outfile_name: filename of file to be written
            payload: the dict information to write to file
        """
        logger.debug(f"FileManager: started writing file '{outfile_name}.json'.")

        with open(
            file=self.outdir.joinpath(f"{outfile_name}.json"),
            mode="w",
            encoding="utf-8",
        ) as outfile:
            outfile.write(json.dumps(payload, indent=4, ensure_ascii=False))

        logger.debug(f"FileManager: completed writing file '{outfile_name}.json'.")



"""Parsing of data from mite json input format

Copyright (c) 2024 to present Mitja Maximilian Zdouc, PhD and individual contributors.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging
from typing import Any, Self

from pydantic import BaseModel

from mite_extras.processing.data_classes import (
    Changelog,
    Cofactors,
    Entry,
    EnyzmeDatabaseIds,
    Enzyme,
    EnzymeAux,
    Evidence,
    Reaction,
    ReactionDatabaseIds,
    ReactionEx,
)

logger = logging.getLogger("mite_extras")


class MiteParser(BaseModel):
    """Assign data from mite input files to internal data structure.

    Attributes:
        entry: an Entry object to represent MITE data
    """

    entry: Any | None = None

    def to_json(self: Self) -> dict:
        """Prepare for export to json

        Returns:
            A dict formatted to be written to json

        Raises:
            RuntimeError: entry has not been assigned yet.
        """
        if self.entry:
            return self.entry.to_json()

    def to_html(self: Self) -> dict:
        """Prepare for export to html

        Returns:
            A dict formatted to be written to json
        """
        if self.entry:
            return self.entry.to_html()

    @staticmethod
    def get_cofactors(cofactors: dict | None) -> Cofactors | None:
        """Parse enzyme-related cofactors info

        Args:
            cofactors: a dict with cofactor info

        Returns:
            An Cofactors object or None
        """
        if isinstance(cofactors, dict):
            return Cofactors(**cofactors)

    @staticmethod
    def get_auxenzymes(auxenzymes: list | None) -> list | None:
        """Extract auxiliary enzyme info and converts into internal data structure

        Args:
            auxenzymes: a list of auxiliary enzyme dicts

        Returns:
            A list of EnzymeAux objects or None
        """
        if not auxenzymes:
            return

        return [
            EnzymeAux(
                name=a.get("name"),
                description=a.get("description"),
                databaseIds=EnyzmeDatabaseIds(**a.get("databaseIds")),
            )
            for a in auxenzymes
        ]

    @staticmethod
    def get_databaseids_reaction(
        data: dict | None,
    ) -> ReactionDatabaseIds | None:
        """Parse reaction-related databaseids info

        Args:
            data: a dict with database IDs

        Returns:
            An ReactionDatabaseIds object
        """
        if isinstance(data, dict):
            return ReactionDatabaseIds(**data)

    def get_reactions(self: Self, reactions: list) -> list:
        """Extract reactions infor and converts into internal data structure

        Args:
            reactions: list with reaction data

        Returns:
            A list of Reaction objects
        """
        return [
            Reaction(
                tailoring=rxn.get("tailoring"),
                description=rxn.get("description"),
                reactionSMARTS=rxn.get("reactionSMARTS"),
                reactions=[ReactionEx(**r) for r in rxn.get("reactions")],
                evidence=Evidence(
                    evidenceCode=rxn.get("evidence", {}).get("evidenceCode"),
                    references=rxn.get("evidence", {}).get("references"),
                ),
                databaseIds=self.get_databaseids_reaction(rxn.get("databaseIds")),
            )
            for rxn in reactions
        ]

    def parse_mite_json(self: Self, data: dict):
        """Parse data from mite-formatted json dict

        Args:
            data: a dict following the mite json schema
        """

        logger.debug("MiteParser: started creating Entry object.")

        try:
            self.entry = Entry(
                accession=data.get("accession"),
                status=data.get("status"),
                retirementReasons=data.get("retirementReasons"),
                changelog=[Changelog(**val) for val in data["changelog"]],
                enzyme=Enzyme(
                    name=data["enzyme"]["name"],
                    description=data.get("enzyme", {}).get("description"),
                    databaseIds=EnyzmeDatabaseIds(**data["enzyme"]["databaseIds"]),
                    auxiliaryEnzymes=self.get_auxenzymes(
                        auxenzymes=data.get("enzyme", {}).get("auxiliaryEnzymes")
                    ),
                    references=data.get("enzyme", {}).get("references"),
                    cofactors=self.get_cofactors(
                        cofactors=data.get("enzyme", {}).get("cofactors")
                    ),
                ),
                reactions=self.get_reactions(reactions=data.get("reactions")),
                comment=data.get("comment"),
            )
        except Exception as e:
            msg = str(e).split("For further information visit")[0].rstrip()
            msg = f"Error in parsing entry {data.get("accession")}: {msg}"
            raise ValueError(msg) from e

        logger.debug("MiteParser: completed creating Entry object.")



"""Validation functionality.

Copyright (c) 2024 to present Mitja Maximilian Zdouc, PhD and Adriano
Rutz (0000-0003-0443-9902) and individual contributors.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import logging
import re
from itertools import permutations, product
from math import pi

import requests
from pydantic import BaseModel
from rdkit.Chem import (
    CanonSmiles,
    GetMolFrags,
    Mol,
    MolFromSmarts,
    MolFromSmiles,
    MolToSmarts,
    MolToSmiles,
    rdMolEnumerator,
)
from rdkit.Chem.rdChemReactions import ChemicalReaction, ReactionFromSmarts

logger = logging.getLogger("mite_extras")


class IdValidator(BaseModel):
    """Handles validation and cross-referencing of database identifiers."""

    def cleanup_ids(
        self, genpept: str | None = None, uniprot: str | None = None
    ) -> dict[str, str]:
        """Cleans up IDs using the UniProt SPARQL endpoint.

        Args:
            genpept: An EMBL ID to be converted to UniProt ID.
            uniprot: A UniProt ID to be converted to EMBL ID.

        Returns:
            A dictionary with the original and corresponding ID.

        Raises:
            ValueError: If neither genpept nor uniprot is provided, or if the IDs do not match.
        """
        if not genpept and not uniprot:
            raise ValueError("Please provide one of NCBI Genpept or Uniprot IDs.")

        def fetch_result(query: str) -> str:
            response = requests.get(
                "https://sparql.uniprot.org/sparql",
                params={"query": query, "format": "srj"},
                headers={"Accept": "application/sparql-results+json"},
                timeout=pi,
            )
            if not response.ok:
                raise ValueError(
                    f"HTTP Error while querying Uniprot: {response.status_code}"
                )

            response_json = response.json()
            bindings = response_json.get("results", {}).get("bindings", [])

            if not bindings:
                raise ValueError("No results found in Uniprot response")

            protein_data = bindings[0].get("protein")
            if not protein_data or "value" not in protein_data:
                raise ValueError(
                    "'protein' key or its 'value' is missing in the Uniprot response"
                )

            protein_uri = protein_data["value"]
            return protein_uri.rsplit("/", 1)[-1]

        def build_genpept_query(genpept: str) -> str:
            return f"""
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            PREFIX up: <http://purl.uniprot.org/core/>
            SELECT ?protein
            WHERE {{
                VALUES ?target {{<http://purl.uniprot.org/embl-cds/{genpept}>}} 
                ?protein a up:Protein .
                ?protein rdfs:seeAlso ?target.
            }}
            """

        def build_uniprot_query(uniprot: str) -> str:
            return f"""
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            PREFIX up: <http://purl.uniprot.org/core/>
            SELECT ?protein
            WHERE {{
                VALUES ?prefix {{<http://purl.uniprot.org/database/EMBL>}}
                <http://purl.uniprot.org/uniprot/{uniprot}> rdfs:seeAlso ?protein .
                ?protein up:database ?prefix .
            }}
            """

        try:
            genpept_result = None
            uniprot_result = None

            if genpept:
                genpept_result = fetch_result(build_genpept_query(genpept))
            if uniprot:
                uniprot_result = fetch_result(build_uniprot_query(uniprot))

            if genpept and uniprot and genpept_result != uniprot:
                raise ValueError(
                    f"The provided Genpept ID '{genpept}' and Uniprot ID '{uniprot}' do not correspond to each other!"
                )

            return {
                "genpept": genpept if genpept else uniprot_result,
                "uniprot": uniprot if uniprot else genpept_result,
            }

        except Exception as e:
            raise ValueError(f"Error during enzyme ID validation: {e!s}") from e

    @staticmethod
    def validate_wikidata_qid(qid: str) -> None:
        """Checks if Wikidata QID exists

        Args:
            qid: a valid Wikidata QID

        Raises:
            ValueError: Wikidata QUI does not exist
        """

        def build_query(qid: str) -> str:
            return f"""
            ASK {{
                wd:{qid} ?p ?o
            }}
            """

        def fetch_result(query: str) -> str:
            response = requests.get(
                "https://query.wikidata.org/sparql",
                params={"query": query},
                headers={"Accept": "application/sparql-results+json"},
                timeout=pi,
            )
            if not response.ok:
                raise ValueError(
                    f"HTTP Error while querying Wikidata: {response.status_code}"
                )
            data = response.json()
            return data.get("boolean")

        if not fetch_result(query=build_query(qid=qid)):
            raise ValueError(
                f"Wikidata QID '{qid}' does not exist or has no statements."
            )


class MoleculeValidator(BaseModel):
    """Handles basic molecule validation and canonicalization."""

    @staticmethod
    def _clean_string(string: str) -> str:
        """Remove superfluous backslashes and H's from string."""
        string = string.replace("\\\\", "\\")
        string = re.sub(r";h\d", "", string)
        return string

    def canonicalize_smiles(self, smiles: str) -> str:
        """Canonicalize a SMILES string."""
        mol = MolFromSmiles(self._clean_string(smiles))
        if mol is None:
            raise ValueError(
                f"RDKit rejected SMILES string - is it a valid SMILES?\n" f"{smiles}"
            )
        for atom in mol.GetAtoms():
            atom.SetAtomMapNum(0)
        return CanonSmiles(MolToSmiles(mol))

    def canonicalize_smarts(self, smarts: str) -> str:
        """Canonicalize a SMARTS pattern."""
        mol = MolFromSmarts(self._clean_string(smarts))
        if mol is None:
            raise ValueError(
                f"RDKit rejected SMARTS string - is it a valid pattern?\n" f"{smarts}"
            )
        for i, atom in enumerate(mol.GetAtoms()):
            atom.SetAtomMapNum(i)
        return MolToSmarts(MolFromSmiles(CanonSmiles(MolToSmiles(mol))))


class ReactionCleaner(BaseModel):
    """Handles cleaning and formatting of reaction SMARTS."""

    @staticmethod
    def clean_ketcher_format(smarts: str) -> str:
        """Clean Ketcher-specific formatting from reaction SMARTS."""
        replacements = {
            # Halogens with indices
            r"-([FCBI]l?):(\d+)": r"-[\1:\2]",
            # Standalone halogens
            r"-([FCBI]l)(?!\[)": r"-[\1]",
            # Substituent halogens with indices
            r"\(-([FCBI]l?):(\d+)\)": r"(-[\1:\2])",
            # Substituent halogens without indices
            r"\(-([FCBI]l)\)": r"(-[\1])",
            # Nitrogen hydrogens in heterocycles
            r"\[#7:(\d+);h(\d+)\]": r"[nH\2:\1]",
            r"\[#7;h(\d+)\]": r"[nH\1]",
            # erroneous specification of charges in indexed atoms
            r"\[(#\d+):(\d+);([+-])\]": r"[\1;\3:\2]",
        }

        for pattern, replacement in replacements.items():
            smarts = re.sub(pattern, replacement, smarts)
        return smarts

    @staticmethod
    def detect_undesired_smarts(smarts: str):
        """Check for undesired pattern in reaction SMARTS

        negative for absence (must not detect)

        Args:
            smarts: reaction SMARTS string

        Raises:
            ValueError: lack of positive pattern/presence of negative pattern
        """

        negative = {
            # explicit hydrogens
            r"\[H\]": "Explicit hydrogen atoms detected in reaction SMARTS (e.g. '[H]'), which is not allowed.\n Depicting chirality? Please specify the stereochemistry using one of the heavy (non-hydrogen) atoms connected to the stereocenter.\n",
            # CXSMARTS
            r"\|": "Reaction SMARTS with CXSMARTS (Chemaxon SMARTS) elements detected, which are not supported by MITE.\nCXSMARTS (aka 'Extended SMARTS') can be recognized with a suffix starting with a pipe character ('|').\n Please export as a Daylight SMARTS or remove the CXSMARTS suffix manually and try again.\n",
        }

        for key, val in negative.items():
            if re.search(key, smarts):
                raise ValueError(f"{val}")


class ReactionEnumerator(BaseModel):
    """Handles enumeration of reactions and molecules."""

    def enumerate_molecule(self, mol: Mol) -> set[Mol]:
        """Enumerate all possible forms of a molecule."""
        if mol is None:
            return set()
        results = rdMolEnumerator.Enumerate(mol)
        return set(results) if results else {mol}

    def enumerate_smarts(self, smarts: str) -> set[str]:
        """Enumerate SMARTS or skip if intramolecular reaction

        Args:
            smarts: a SMARTS string

        Returns:
            A set of SMARTS strings
        """
        if re.match(r"^\(.+\)|\(.+\)$", smarts):
            return {smarts}

        mol = MolFromSmarts(smarts)
        if mol is None:
            raise ValueError(
                f"RDKit rejected SMARTS string - is it a valid pattern?\n" f"{smarts}"
            )
        enumerated_mols = self.enumerate_molecule(mol)
        return {MolToSmarts(m) for m in enumerated_mols if m is not None}

    def generate_smarts_variants(self, smarts: str) -> list[str]:
        """Generate all possible variants of a SMARTS pattern with comma-separated elements."""
        pattern = re.compile(r"\[([^\]:]+(?:,[^\]:]+)*)\:(\d+)\]")
        matches = list(pattern.finditer(smarts))

        if not matches:
            return [smarts]

        options = []
        for m in matches:
            options_list = [f"[{opt}:{m.group(2)}]" for opt in m.group(1).split(",")]
            options.append(options_list)

        variants = []
        for combination in product(*options):
            new_smarts = smarts
            for m, replacement in zip(matches, combination, strict=True):
                new_smarts = new_smarts.replace(m.group(0), replacement, 1)
            variants.append(new_smarts)

        return variants


class ReactionValidator(BaseModel):
    """Main class for validating chemical reactions."""

    molecule_validator: MoleculeValidator = MoleculeValidator()
    reaction_cleaner: ReactionCleaner = ReactionCleaner()
    enumerator: ReactionEnumerator = ReactionEnumerator()

    def validate_reaction(
        self,
        reaction_smarts: str,
        substrate_smiles: str,
        expected_products: list[str],
        forbidden_products: list[str] | None = None,
        intramolecular: bool = False,
    ) -> None:
        """
        Validate a reaction SMARTS against expected and forbidden products.

        Args:
            reaction_smarts: The reaction SMARTS pattern
            substrate_smiles: The substrate SMILES
            expected_products: List of expected product SMILES
            forbidden_products: Optional list of forbidden product SMILES
            intramolecular: Whether the reaction is intramolecular

        Raises:
            ValueError: missing products or overlap expected & forbidden products or not all expected products generated
        """
        if not expected_products:
            raise ValueError("At least one product must be specified")

        self.reaction_cleaner.detect_undesired_smarts(reaction_smarts)

        reaction_smarts = self.reaction_cleaner.clean_ketcher_format(reaction_smarts)

        substrate_smiles = self.molecule_validator._clean_string(substrate_smiles)

        expected_smiles = {
            self.molecule_validator.canonicalize_smiles(p) for p in expected_products
        }
        forbidden_smiles = {
            self.molecule_validator.canonicalize_smiles(p)
            for p in (forbidden_products or [])
        }

        if overlap := forbidden_smiles & expected_smiles:
            raise ValueError(
                f"Overlap between expected and forbidden products:\n"
                f"{'\n'.join(overlap)}\n"
            )

        # Generate reaction variants and validate
        reactions = self._get_reaction_variants(reaction_smarts)
        predicted_products = self._run_reactions(
            reactions, substrate_smiles, intramolecular
        )

        # Validate predictions
        predicted_smiles = {
            self.molecule_validator.canonicalize_smiles(MolToSmiles(p))
            for p in predicted_products
            if p is not None
        }

        if not expected_smiles.issubset(predicted_smiles):
            raise ValueError(
                f"Reaction did not lead to all expected products.\n"
                f"Expected products:\n"
                f"{'\n'.join(expected_smiles)}\n"
                f"Generated products:\n"
                f"{'\n'.join(predicted_smiles)}\n"
            )

        if overlap := forbidden_smiles & predicted_smiles:
            raise ValueError(
                f"Reaction product(s) belong(s) to the specified forbidden product(s):\n"
                f"{'\n'.join(overlap)}\n"
            )

        logger.debug("Successfully validated reaction SMARTS")

    def _get_reaction_variants(self, reaction_smarts: str) -> set[ChemicalReaction]:
        """Get all possible variants of a reaction."""
        try:
            reactants, products = reaction_smarts.split(">>")
        except ValueError as e:
            raise ValueError("Invalid reaction SMARTS format") from e

        reactant_variants = self.enumerator.generate_smarts_variants(reactants)
        product_variants = self.enumerator.generate_smarts_variants(products)

        reactants_list = [
            self.enumerator.enumerate_smarts(reactant) for reactant in reactant_variants
        ]
        products_list = [
            self.enumerator.enumerate_smarts(product) for product in product_variants
        ]

        reactions = set()
        for r in reactants_list:
            for p in products_list:
                reaction_smarts_combinations = [
                    f"{r_smarts}>>{p_smarts}" for r_smarts in r for p_smarts in p
                ]
                for reaction_smarts_combined in reaction_smarts_combinations:
                    reaction = ReactionFromSmarts(reaction_smarts_combined)
                    if reaction is not None:
                        reactions.add(reaction)

        return reactions

    def _run_reactions(
        self,
        reactions: set[ChemicalReaction],
        substrate_smiles: str,
        intramolecular: bool = False,
    ) -> set[Mol]:
        """Run all reaction variants on the substrate."""
        substrate = MolFromSmiles(substrate_smiles)
        if substrate is None:
            raise ValueError(
                f"RDKit rejected SMILES string - is it a valid SMILES?\n"
                f"{substrate_smiles}"
            )

        substrate_variants = self.enumerator.enumerate_molecule(substrate)
        products = set()

        for mol in substrate_variants:
            reactants = GetMolFrags(mol, asMols=True)

            if intramolecular and len(reactants) > 1:
                continue

            reactant_perms = (
                [reactants] if intramolecular else list(permutations(reactants))
            )

            for reaction in reactions:
                for reactant_combo in reactant_perms:
                    try:
                        reaction_products = reaction.RunReactants(reactant_combo)
                        for product_set in reaction_products:
                            products.update(product_set)
                    except Exception as e:
                        logger.debug(f"Error during running of reaction: {e}")
                        continue

        return {p for p in products if p is not None}
